// $Id: Application.cpp,v 1.11 2014/03/12 22:36:57 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Application.h"
#include "Application.moc"

#include "BaseConfigurationDialog.h"
#include "LineEditor.h"
#include "MainWindow.h"
#include "WarningDialog.h"
#include "WindowSelectionFrame.h"

#include <QApplication>

//____________________________________________
Application::Application( CommandLineArguments arguments ) :
    BaseApplication( 0, arguments ),
    Counter( "Application" ),
    mainWindow_(0)
{ Debug::Throw( "Application::Application.\n" ); }

//____________________________________________
Application::~Application( void )
{}

//____________________________________________
bool Application::realizeWidget( void )
{
    Debug::Throw( "Application::realizeWidget.\n" );

    if( !BaseApplication::realizeWidget() ) return false;

    // top widget
    mainWindow_ = new MainWindow();
    mainWindow_->setWindowTitle( applicationName() );

    // configuration
    emit configurationChanged();

    // top widget
    mainWindow_->centerOnDesktop();

    // parse options
    const CommandLineParser parser( commandLineParser( _arguments() ) );
    if( parser.hasOption( "--delay" ) )
    { mainWindow_->windowSelectionFrame().setDelay( parser.option( "--delay" ).toInt() ); }

    if( parser.hasFlag( "--root" ) )
    {

        mainWindow_->windowSelectionFrame().selectRootWindow();
        mainWindow_->show();

    } else if( parser.hasOption( "--wid" ) ) {

        int pos(0);
        QString wIdString( parser.option( "--wid" ) );
        if( mainWindow_->windowSelectionFrame().editor().validator()->validate( wIdString, pos ) != QValidator::Acceptable )
        {

            mainWindow_->show();
            WarningDialog( mainWindow_, QString( tr( "'%1' is not a valid window identifier." ) ).arg( wIdString ) ).exec();

        } else {

            mainWindow_->windowSelectionFrame().editor().setText( wIdString );
            mainWindow_->windowSelectionFrame().selectWindowFromEditor();
            mainWindow_->show();

        }

    } else if( parser.hasFlag( "--grab" ) ) {

        if( !XmlOptions::get().get<bool>( "AUTO_HIDE" ) ) mainWindow_->show();
        if( parser.hasOption( "--delay" ) ) mainWindow_->windowSelectionFrame().selectWindowFromPointer();
        else mainWindow_->windowSelectionFrame().selectWindowFromPointer( true );

    } else mainWindow_->show();

    return true;

}

//____________________________________________
CommandLineParser Application::commandLineParser( CommandLineArguments arguments, bool ignoreWarnings ) const
{
    Debug::Throw( "Application::commandLineParser.\n" );
    CommandLineParser out( BaseCoreApplication::commandLineParser() );
    out.setGroup( CommandLineParser::applicationGroupName );
    out.registerFlag( "--root", tr( "examine the root window" ) );
    out.registerFlag( "--grab", tr( "select a window using mouse pointer before starting the application" ) );
    out.registerOption( "--delay", tr( "integer" ), tr( "delay (in seconds) before examining the window located below the mouse pointer" ) );
    out.registerOption( "--wid", tr( "integer" ), tr( "resource id of window to examine" ) );
    if( !arguments.isEmpty() ) out.parse( arguments, ignoreWarnings );
    return out;
}

//____________________________________________
void Application::usage( void ) const
{
    _usage( "xgrabwindow" );
    commandLineParser().usage();
    return;
}

//_______________________________________________
void Application::_configuration( void )
{

    Debug::Throw( "Application::_configuration.\n" );
    emit saveConfiguration();

    BaseConfigurationDialog dialog( 0, BaseConfigurationDialog::Default );
    dialog.setWindowTitle( QString( tr( "Configuration - %1" ) ).arg( applicationName() ) );

    dialog.centerOnWidget( qApp->activeWindow() );
    dialog.exec();

}

//________________________________________________
bool Application::_processCommand( Server::ServerCommand command )
{

    Debug::Throw( "Application::_processCommand.\n" );
    if( BaseApplication::_processCommand( command ) ) return true;
    if( command.command() == Server::ServerCommand::Raise )
    {

        mainWindow_->uniconify();

        CommandLineParser parser( commandLineParser( command.arguments() ) );
        if( parser.hasOption( "--delay" ) )
        { mainWindow_->windowSelectionFrame().setDelay( parser.option( "--delay" ).toInt() ); }

        if( parser.hasFlag( "--root" ) )
        {

            mainWindow_->windowSelectionFrame().selectRootWindow();

        } else if( parser.hasOption( "--wid" ) ) {

            int pos(0);
            QString wIdString( parser.option( "--wid" ) );
            if( mainWindow_->windowSelectionFrame().editor().validator()->validate( wIdString, pos ) != QValidator::Acceptable )
            {

                WarningDialog( mainWindow_, QString( tr( "'%1' is not a valid window identifier." ) ).arg( wIdString ) ).exec();

            } else {

                mainWindow_->windowSelectionFrame().editor().setText( wIdString );
                mainWindow_->windowSelectionFrame().selectWindowFromEditor();

            }

        } else if( parser.hasFlag( "--grab" ) ) {

            mainWindow_->windowSelectionFrame().selectWindow();

        }

        return true;

    } else return false;

}
