// $Id: MainWindow.cpp,v 1.22 2014/05/09 07:21:28 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "MainWindow.h"
#include "MainWindow.moc"

#include "Application.h"
#include "AnimatedStackedWidget.h"
#include "AnimatedTreeView.h"
#include "BaseStatusBar.h"
#include "ExistingFilesDialog.h"
#include "FileDialog.h"
#include "FileRecord.h"
#include "IconNames.h"
#include "IconEngine.h"
#include "Menu.h"
#include "Singleton.h"
#include "WindowCaptureDialog.h"
#include "WindowCaptureWidget.h"
#include "WindowPropertiesDisplay.h"
#include "WindowSelectionFrame.h"
#include "WindowTreeWidget.h"

#include "XcbUtil.h"
#include "XmlOptions.h"

#include <QLayout>
#include <QSplitter>

//_____________________________________________________________
MainWindow::MainWindow( QWidget* parent ):
    BaseMainWindow( parent )
{
    Debug::Throw( "MainWindow::MainWindow.\n" );
    setOptionName( "MAIN_WINDOW" );

    // central widget
    QSplitter* splitter( new QSplitter( this ) );
    setCentralWidget( splitter );
    splitter->setOrientation( Qt::Horizontal );

    // window tree
    windowTreeWidget_ = new WindowTreeWidget();

    // default width from options, if found
    if( XmlOptions::get().contains( "WINDOW_TREE_WIDTH" ) )
    { windowTreeWidget_->setDefaultWidth( XmlOptions::get().get<int>( "WINDOW_TREE_WIDTH" ) ); }

    QWidget* mainWidget = new QWidget();
    QVBoxLayout* vLayout( new QVBoxLayout() );
    vLayout->setMargin(0);
    vLayout->setSpacing(5);
    mainWidget->setLayout( vLayout );

    // right widget
    stackedWidget_ = new AnimatedStackedWidget( mainWidget );
    vLayout->addWidget( stackedWidget_, 1 );

    // window properties
    windowPropertiesWidget_ = new QWidget();
    stackedWidget_->addWidget( windowPropertiesWidget_ );
    windowPropertiesWidget_->setLayout( new QVBoxLayout() );
    windowPropertiesWidget_->layout()->setMargin(0);
    windowPropertiesWidget_->layout()->setSpacing(5);

    // actions
    Application& application( *Singleton::get().application<Application>() );
    addAction( &application.closeAction() );

    // properties display
    reinterpret_cast<QVBoxLayout*>(windowPropertiesWidget_->layout())->addWidget( windowPropertiesDisplay_ = new WindowPropertiesDisplay( windowPropertiesWidget_ ), 0 );
    windowPropertiesDisplay_->clear();

    // properties list
    reinterpret_cast<QVBoxLayout*>(windowPropertiesWidget_->layout())->addWidget( list_ = new AnimatedTreeView( windowPropertiesWidget_ ), 1 );
    list_->setModel( &model_ );
    list_->setOptionName( "ATOM_DATA_LIST" );

    // clear widget
    QLabel* label = new QLabel( tr( "No Window Selected" ) );
    label->setAlignment( Qt::AlignCenter );
    clearWidget_ = label;
    stackedWidget_->addWidget( clearWidget_ );

    // create selection frame
    vLayout->addWidget( windowSelectionFrame_ = new WindowSelectionFrame( mainWidget ), 0 );
    windowSelectionFrame_->selectWindowButton().setFocus();

    // actions
    _installActions();

    // status bar
    BaseStatusBar* statusBar;
    setStatusBar( statusBar = new BaseStatusBar( this ) );
    statusBar->addLabel( 1, true );
    statusBar->addClock();

    // populate splitter
    splitter->addWidget( windowTreeWidget_ );
    splitter->addWidget( mainWidget );
    splitter->setStretchFactor( 0, 0 );
    splitter->setStretchFactor( 1, 1 );
    connect( splitter, SIGNAL(splitterMoved(int,int)), SLOT(_splitterMoved()) );

    // show clearWidget
    stackedWidget_->setCurrentWidget( clearWidget_ );

    // window capture widget
    windowCaptureDialog_ = new WindowCaptureDialog();
    connect( windowCaptureDialog_, SIGNAL(dialogHidden()), SLOT(_windowCaptureDialogHidden()) );

    // menu bar
    setMenuBar( menu_ = new Menu( this ) );

    // connections
    connect( windowSelectionFrame_, SIGNAL(windowSelectionChanged()), SLOT(_updateSelectedWindow()));
    connect( windowTreeWidget_, SIGNAL(windowActivated(WId)), windowSelectionFrame_, SLOT(selectWindowFromTree(WId)) );

    // configuration
    connect( &application, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();

    // xcb thread
    connect( &xcbThread_, SIGNAL(windowCreated(WId)), windowTreeWidget_, SLOT(reloadOnWindowChanged()) );
    connect( &xcbThread_, SIGNAL(windowDestroyed(WId)), windowTreeWidget_, SLOT(reloadOnWindowChanged()) );
    xcbThread_.start();

}

//_______________________________________________________
MainWindow::~MainWindow( void )
{}

//_______________________________________________________
void MainWindow::timerEvent( QTimerEvent* event )
{

    if( event->timerId() == resizeTimer_.timerId() )
    {

        // stop timer and save time
        resizeTimer_.stop();
        if( windowTreeWidget_->visibilityAction().isChecked() )
        { XmlOptions::get().set<int>( "WINDOW_TREE_WIDTH", windowTreeWidget_->width() ); }

    } else return BaseMainWindow::timerEvent( event );

}

//____________________________________
void MainWindow::closeEvent( QCloseEvent *event )
{
    Debug::Throw( "MainWindow::closeEvent.\n" );
    event->accept();
    qApp->quit();
}

//_____________________________________________________________
void MainWindow::_updateSelectedWindow( void )
{
    Debug::Throw( "MainWindow::_updateSelectedWindow.\n" );
    const WindowProperties& properties( windowSelectionFrame_->windowProperties() );
    windowPropertiesDisplay_->setWindowProperties( properties );
    stackedWidget_->setCurrentWidget( properties.isValid() ? windowPropertiesWidget_ : clearWidget_ );

    // update model
    AtomDataModel::List atoms;
    const AtomData::Hash& atomHash( properties.atoms() );
    for( AtomData::Hash::const_iterator iter = atomHash.constBegin(); iter != atomHash.constEnd(); ++iter )
    { atoms.append( iter.value() ); }

    model_.set( atoms );
    list_->resizeColumns();

    saveIconAction_->setEnabled( !windowPropertiesDisplay_->icon().isNull() );
    reloadAction_->setEnabled( windowSelectionFrame_->selectedWindow() );
    clearAction_->setEnabled( windowSelectionFrame_->selectedWindow() );

    // create window tree item and select in tree
    WindowTreeItem item( properties.windowId(), properties.parentId() );
    windowTreeWidget_->selectWindow( item );

    // update window capture widget
    windowCaptureDialog_->windowCaptureWidget().setWindowProperties( properties );
    windowCaptureDialog_->reloadButton().setEnabled( properties.isValid() );

    // update label
    if( windowCaptureDialog_->isVisible() && properties.isValid() )
    {
        windowCaptureDialog_->windowCaptureWidget().updateLabelPixmap();
        windowCaptureDialog_->raise();
    }

}

//_____________________________________________________________
void MainWindow::_clear( void )
{
    Debug::Throw( "MainWindow::_clear.\n" );
    windowPropertiesDisplay_->clear();
    model_.clear();
    stackedWidget_->setCurrentWidget( clearWidget_ );

    clearAction_->setEnabled(false);
    reloadAction_->setEnabled(false);
    saveIconAction_->setEnabled(false);

}

//_____________________________________________________________
void MainWindow::_saveIcon( void )
{
    Debug::Throw( "MainWindow::_saveIcon.\n" );
    const QIcon icon( windowPropertiesDisplay_->icon() );
    if( icon.isNull() )
    { return; }

    // generate default file name
    QString fileName;
    QString windowClass( windowPropertiesDisplay_->windowClass() );
    if( !windowClass.isEmpty() )
    {

        fileName = windowClass.left( windowClass.indexOf( QRegExp( "\\s" ) ) );
        fileName += ".png";

    } else fileName = "icon.png";

    FileDialog dialog( this );
    dialog.setAcceptMode( QFileDialog::AcceptSave );
    dialog.setFileMode( QFileDialog::AnyFile );
    dialog.selectFile( File( fileName ).addPath( dialog.workingDirectory() ) );

    //
    File file( dialog.getFile() );
    if( file.isNull() ) return;
    else file = file.expand();

    const QList<QSize> sizes( icon.availableSizes() );
    if( sizes.size() == 1 )
    {

        icon.pixmap( sizes[0] ).save( file );

    } else {

        FileRecord::List existingIconFiles;
        foreach( const QSize& size, sizes )
        {

            File iconFile( QString( "%1-%2x%3.%4" ).arg( file.truncatedName() ).arg( size.width() ).arg( size.height() ).arg( file.extension() ) );
            if( iconFile.exists() ) existingIconFiles.append( iconFile );

        }

        if( existingIconFiles.empty() || ExistingFilesDialog( this, existingIconFiles ).exec() )
        {
            foreach( const QSize& size, sizes )
            { icon.pixmap( size ).save( QString( "%1-%2x%3.%4" ).arg( file.truncatedName() ).arg( size.width() ).arg( size.height() ).arg( file.extension() ) ); }

        }

    }

    return;

}

//________________________________________________________
void MainWindow::_splitterMoved( void )
{
    Debug::Throw( "MainWindow::_splitterMoved.\n" );
    resizeTimer_.start( 200, this );
}

//_______________________________________________
void MainWindow::_showWindowCaptureDialog( void )
{
    Debug::Throw( "MainWindow::_showWindowCaptureDialog.\n" );
    windowCaptureDialog_->show();
    windowCaptureDialog_->raise();
    XmlOptions::get().set<bool>( "SHOW_WINDOW_CAPTURE", true );
}

//_______________________________________________
void MainWindow::_windowCaptureDialogHidden( void )
{
    Debug::Throw( "MainWindow::_windowCaptureDialogHidden.\n" );
    XmlOptions::get().set<bool>( "SHOW_WINDOW_CAPTURE", false );
}

//____________________________________________________________________________________________
void MainWindow::_updateConfiguration( void )
{
    Debug::Throw( "MainWindow::_updateConfiguration.\n" );

    // log and command list
    if( XmlOptions::get().get<bool>( "SHOW_WINDOW_CAPTURE" ) ) _showWindowCaptureDialog();
    else windowCaptureDialog_->hide();

}

//_____________________________________________________________
void MainWindow::_installActions( void )
{
    Debug::Throw( "MainWindow::_installActions.\n" );

    addAction( clearAction_ = new QAction( IconEngine::get( IconNames::Clear ), tr( "Clear" ), this ) );
    clearAction_->setToolTip( tr( "Clear loaded window properties" ) );
    connect( clearAction_, SIGNAL(triggered()), SLOT(_clear()) );
    clearAction_->setEnabled( false );

    addAction( reloadAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Reload Window Properties" ), this ) );
    reloadAction_->setToolTip( tr( "Reload properties of the current window" ) );
    reloadAction_->setShortcut( QKeySequence::Refresh );
    connect( reloadAction_, SIGNAL(triggered()), windowSelectionFrame_, SLOT( reload()) );
    reloadAction_->setEnabled( false );

    addAction( saveIconAction_ = new QAction( IconEngine::get( IconNames::SaveAs ), tr( "Save Icon ..." ), this ) );
    saveIconAction_->setShortcut( QKeySequence::SaveAs );
    saveIconAction_->setToolTip( tr( "Save application's icon to disk." ) );
    connect( saveIconAction_, SIGNAL(triggered()), SLOT(_saveIcon()) );
    saveIconAction_->setEnabled( false );

    addAction( windowCaptureAction_ = new QAction( tr( "Show Window Capture" ), this ) );
    connect( windowCaptureAction_, SIGNAL(triggered()), SLOT(_showWindowCaptureDialog()) );

}
