// $Id: WindowCaptureWidget.cpp,v 1.3 2014/05/08 11:43:57 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "WindowCaptureWidget.h"
#include "WindowCaptureWidget_p.h"
#include "WindowCaptureWidget.moc"

#include "IconEngine.h"
#include "IconNames.h"
#include "Singleton.h"
#include "XcbWindowCapture.h"
#include "XmlOptions.h"

#include <QAction>
#include <QLayout>
#include <QPainter>

//____________________________________________________________
WindowCaptureWidget::WindowCaptureWidget( QWidget* parent ):
    QWidget( parent ),
    Counter( "WindowCaptureWidget" ),
    shadowSize_( 12 ),
    overlap_( 3 ),
    shadowTiles_( 0x0 )
{

    Debug::Throw( "WindowCaptureWidget::WindowCaptureWidget.\n" );
    QVBoxLayout* layout( new QVBoxLayout() );
    layout->setMargin(shadowSize_ - overlap_ );
    layout->setSpacing(5);
    setLayout( layout );

    layout->addWidget( label_ = new QLabel( this ) );
    label_->setAlignment( Qt::AlignCenter );
    label_->setText( tr( "No Window Selected" ) );
    label_->setSizePolicy( QSizePolicy::Ignored, QSizePolicy::Ignored );
    _createShadowTiles();

}

//____________________________________________________________
WindowCaptureWidget::~WindowCaptureWidget( void )
{ delete shadowTiles_; }

//____________________________________________________________
void WindowCaptureWidget::setWindowProperties( const WindowProperties& properties )
{
    Debug::Throw( "WindowCaptureWidget::setWindowProperties.\n" );
    if( properties.isValid() )
    {

        window_ = properties.windowId();
        reload();

    } else {

        window_ = 0x0;
        pixmap_ = QPixmap();
        label_->setText( tr( "No Window Selected" ) );

    }

}

//____________________________________________________________
void WindowCaptureWidget::updateLabelPixmap( void )
{
    Debug::Throw( "WindowCaptureWidget::updateLabelPixmap.\n" );
    if( !pixmap_.isNull() ) label_->setPixmap( pixmap_.scaled( label_->size(), Qt::KeepAspectRatio, Qt::SmoothTransformation ) );
    update();
}

//____________________________________________________________
void WindowCaptureWidget::reload( void )
{
    Debug::Throw( "WindowCaptureWidget::reload.\n" );

    // check window
    if( !window_ ) return;

    // capture and update
    pixmap_ = XcbWindowCapture( this ).capture( window_ );
    if( pixmap_.isNull() ) label_->setText( tr( "Captured Image is Invalid" ) );
    else updateLabelPixmap();

}

//____________________________________________________________
void WindowCaptureWidget::resizeEvent( QResizeEvent* event )
{
    QWidget::resizeEvent( event );
    timer_.start( 200, this );
    return;
}

//____________________________________________________________
void WindowCaptureWidget::timerEvent( QTimerEvent* event )
{
    Debug::Throw( "WindowCaptureWidget::timerEvent.\n" );
    if( event->timerId() == timer_.timerId() )
    {
        timer_.stop();
        updateLabelPixmap();
    } else return QWidget::timerEvent( event );
}

//____________________________________________________________
void WindowCaptureWidget::paintEvent( QPaintEvent* event )
{

    if( !pixmap_.isNull() )
    {
        QPainter p( this );
        p.setClipRegion( event->region() );

        // calculate proper rect
        const int margin( shadowSize_ - overlap_ );
        const QRect pixmapRect( label_->pixmap()->rect().intersected( label_->rect() ).adjusted( -margin, -margin, margin, margin )  );
        const QRect rect( QPoint(
            ( this->rect().width() - pixmapRect.width())/ 2,
            ( this->rect().height() - pixmapRect.height())/ 2 ),
            pixmapRect.size() );

        shadowTiles_->render( rect, &p );
    }
    return;

}

//____________________________________________________________
void WindowCaptureWidget::_createShadowTiles( void )
{
    Debug::Throw( "WindowCaptureWidget::_createShadowTiles.\n" );

    const qreal fixedSize( 25.5 );
    const qreal size( shadowSize_ );
    const qreal verticalOffset( 0.2 );

    QPixmap shadow( size*2, size*2 );
    shadow.fill( Qt::transparent );
    QPainter p( &shadow );
    p.setRenderHint( QPainter::Antialiasing );
    p.setPen( Qt::NoPen );

    {
        // inner (sharp gradient)
        const qreal gradientSize = qMin( size, fixedSize );
        const qreal voffset( 0.2 );

        QRadialGradient rg = QRadialGradient( size, size+voffset, gradientSize );
        rg.setColorAt(1, Qt::transparent );

        // parabolic shadow is used
        int nPoints( (10*gradientSize)/fixedSize );
        Private::Parabolic f( 1.0, 0.22 );
        QColor c( Qt::black );
        for( int i = 0; i < nPoints; i++ )
        {
            qreal x = qreal(i)/nPoints;
            c.setAlphaF( f(x) );
            rg.setColorAt( x, c );

        }

        p.setBrush( rg );
        p.drawRect( shadow.rect() );

    }

    {

        // mid gradient
        const qreal gradientSize = qMin( size, (size+2*fixedSize)/3 );
        const qreal voffset = qMin( 8.0*verticalOffset/fixedSize, 4.0 );

        // gaussian shadow is used
        QRadialGradient rg = QRadialGradient( size, size+voffset, gradientSize );
        rg.setColorAt(1, Qt::transparent );

        int nPoints( (10*gradientSize)/fixedSize );
        Private::Gaussian f( 0.54, 0.21);
        QColor c( Qt::black );
        for( int i = 0; i < nPoints; i++ )
        {
            qreal x = qreal(i)/nPoints;
            c.setAlphaF( f(x) );
            rg.setColorAt( x, c );

        }

        p.setBrush( rg );
        p.drawRect( shadow.rect() );

    }

    {

        // outer (spread) gradient
        const qreal gradientSize = size;
        const qreal voffset = qMin( 20.0*verticalOffset/fixedSize, 4.0 );

        // gaussian shadow is used
        QRadialGradient rg = QRadialGradient( size, size+voffset, gradientSize );
        rg.setColorAt(1, Qt::transparent );

        int nPoints( (20*gradientSize)/fixedSize );
        Private::Gaussian f( 0.155, 0.445);
        QColor c( Qt::black );
        for( int i = 0; i < nPoints; i++ )
        {
            qreal x = qreal(i)/nPoints;
            c.setAlphaF( f(x) );
            rg.setColorAt( x, c );
        }

        p.setBrush( rg );
        p.drawRect( shadow.rect() );

    }

    p.setCompositionMode(QPainter::CompositionMode_DestinationOut);
    p.setBrush( Qt::black );
    p.drawEllipse( QRectF( size-overlap_+1, size-overlap_+1, 2*(overlap_-1), 2*(overlap_-1) ) );

    // split shadow into tiles
    shadowTiles_ = new Private::TileSet(shadow, size, size, 1, 1);

}

namespace Private
{

     //______________________________________________________________
    int TileSet::_sideExtent = 32;

    //______________________________________________________________
    void TileSet::initPixmap( PixmapList& pixmaps, const QPixmap &pix, int w, int h, const QRect &rect)
    {
        QSize size( w, h );
        if( !( size.isValid() && rect.isValid() ) )
        {
            pixmaps.append( QPixmap() );

        } else if( size != rect.size() ) {

            const QPixmap tile( pix.copy(rect) );
            QPixmap pixmap( w, h );

            pixmap.fill(Qt::transparent);
            QPainter p(&pixmap);
            p.drawTiledPixmap(0, 0, w, h, tile);

            pixmaps.append( pixmap );

        } else pixmaps.append( pix.copy(rect) );

    }

    //______________________________________________________________
    TileSet::TileSet(const QPixmap &pix, int w1, int h1, int w2, int h2 ):
        _w1(w1),
        _h1(h1),
        _w3(0),
        _h3(0)
    {
        _pixmaps.reserve(9);
        if (pix.isNull()) return;

        _w3 = pix.width() - (w1 + w2);
        _h3 = pix.height() - (h1 + h2);
        int w = w2;
        int h = h2;
        {
            while (w < _sideExtent && w2 > 0) w += w2;
            while (h < _sideExtent && h2 > 0) h += h2;
        }

        // initialise pixmap array
        initPixmap( _pixmaps, pix, _w1, _h1, QRect(0, 0, _w1, _h1) );
        initPixmap( _pixmaps, pix, w, _h1, QRect(_w1, 0, w2, _h1) );
        initPixmap( _pixmaps, pix, _w3, _h1, QRect(_w1+w2, 0, _w3, _h1) );
        initPixmap( _pixmaps, pix, _w1, h, QRect(0, _h1, _w1, h2) );
        initPixmap( _pixmaps, pix, w, h, QRect(_w1, _h1, w2, h2) );
        initPixmap( _pixmaps, pix, _w3, h, QRect(_w1+w2, _h1, _w3, h2) );
        initPixmap( _pixmaps, pix, _w1, _h3, QRect(0, _h1+h2, _w1, _h3) );
        initPixmap( _pixmaps, pix, w, _h3, QRect(_w1, _h1+h2, w2, _h3) );
        initPixmap( _pixmaps, pix, _w3, _h3, QRect(_w1+w2, _h1+h2, _w3, _h3) );
    }

    //___________________________________________________________
    void TileSet::render(const QRect &r, QPainter *p) const
    {

        // check initialization
        if( _pixmaps.size() < 9 ) return;

        int x0, y0, w, h;
        r.getRect(&x0, &y0, &w, &h);

        // calculate pixmaps widths
        int wLeft(0);
        int wRight(0);
        if( _w1+_w3 > 0 )
        {
            qreal wRatio( qreal( _w1 )/qreal( _w1 + _w3 ) );
            wLeft = qMin( _w1, int(w*wRatio) );
            wRight = qMin( _w3, int(w*(1.0-wRatio)) );
        }

        // calculate pixmap heights
        int hTop(0);
        int hBottom(0);
        if( _h1+_h3 > 0 )
        {
            qreal hRatio( qreal( _h1 )/qreal( _h1 + _h3 ) );
            hTop = qMin( _h1, int(h*hRatio) );
            hBottom = qMin( _h3, int(h*(1.0-hRatio)) );
        }

        // calculate corner locations
        w -= wLeft + wRight;
        h -= hTop + hBottom;
        const int x1 = x0 + wLeft;
        const int x2 = x1 + w;
        const int y1 = y0 + hTop;
        const int y2 = y1 + h;

        // corner
        p->drawPixmap(x0, y0, _pixmaps.at(0), 0, 0, wLeft, hTop);
        p->drawPixmap(x2, y0, _pixmaps.at(2), _w3-wRight, 0, wRight, hTop);
        p->drawPixmap(x0, y2, _pixmaps.at(6), 0, _h3-hBottom, wLeft,  hBottom);
        p->drawPixmap(x2, y2, _pixmaps.at(8), _w3-wRight, _h3-hBottom, wRight, hBottom );

        // top and bottom
        if( w > 0 )
        {
            p->drawTiledPixmap(x1, y0, w, hTop, _pixmaps.at(1));
            p->drawTiledPixmap(x1, y2, w, hBottom, _pixmaps.at(7), 0, _h3-hBottom );
        }

        // left and right
        if( h > 0 )
        {
            p->drawTiledPixmap(x0, y1, wLeft, h, _pixmaps.at(3));
            p->drawTiledPixmap(x2, y1, wRight, h, _pixmaps.at(5), _w3-wRight, 0 );
        }

        // center
        p->drawTiledPixmap(x1, y1, w, h, _pixmaps.at(4));

    }

}
