// $Id: WindowProperties.cpp,v 1.16 2014/05/08 10:15:38 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "WindowProperties.h"

#include "AtomData.h"
#include "Debug.h"
#include "XcbProperties.h"
#include "XcbUtil.h"

#include <string>

//_____________________________________________________________________
WindowProperties::WindowProperties( void ):
    Counter( "WindowProperties" ),
    valid_( false ),
    window_( 0x0 ),
    parent_( 0x0 )
{}

//_____________________________________________________________________
WindowProperties::WindowProperties( WId window, XcbProperties* xcbProperties ):
    Counter( "WindowProperties" ),
    valid_( false ),
    window_( window ),
    parent_( 0x0 )
{

    Debug::Throw() << "WindowProperties::WindowProperties - window: " << XcbUtil::get().windowIdString( window ) << endl;

    // do nothing if window is invalid
    if( !window_ ) return;

    // get all properties
    #if HAVE_X11

    xcb_connection_t* connection( XcbUtil::get().connection<xcb_connection_t>() );

    {
        // find all atoms
        xcb_list_properties_cookie_t cookie( xcb_list_properties( connection, window ) );
        XcbUtil::ScopedPointer<xcb_list_properties_reply_t> reply( xcb_list_properties_reply( connection, cookie, 0x0 ) );

        if( !reply ) return;

        xcb_atom_t* atoms( xcb_list_properties_atoms( reply.data() ) );
        int count( xcb_list_properties_atoms_length( reply.data() ) );

        // loop over properties
        for( int i=0; i<count; ++i )
        {

            //! atom data and type
            AtomData::Pointer atomData( xcbProperties->atomData( window, atoms[i] ) );

            Debug::Throw() << "WindowProperties::WindowProperties -"
                << " atom: " << atomData->name() << " (" << atomData->type() << "): "
                << atomData->stringValue()
                << endl;

            atoms_.insert( atomData->name(), atomData );

        }

    }

    {

        // geometry
        xcb_get_geometry_cookie_t cookie( xcb_get_geometry( connection, window ) );
        XcbUtil::ScopedPointer<xcb_get_geometry_reply_t> reply( xcb_get_geometry_reply( connection, cookie, 0x0 ) );
        if( reply )
        {
            size_ = QSize( reply.data()->width, reply.data()->height );

            // translate coordinates
            xcb_translate_coordinates_cookie_t coordCookie( xcb_translate_coordinates(
                connection, window, reply.data()->root,
                -reply.data()->border_width,
                -reply.data()->border_width ) );

            XcbUtil::ScopedPointer< xcb_translate_coordinates_reply_t> coordReply( xcb_translate_coordinates_reply(
                connection, coordCookie, 0x0 ) );

            if( coordReply )
            {
                position_.setX( coordReply.data()->dst_x );
                position_.setY( coordReply.data()->dst_y );
            }

        }

    }

    // parents
    QVector<xcb_window_t> parents;
    {
        WId current( window_ );
        while( true )
        {
            xcb_query_tree_cookie_t cookie = xcb_query_tree( connection, current );
            XcbUtil::ScopedPointer<xcb_query_tree_reply_t> reply(xcb_query_tree_reply( connection, cookie, 0x0 ) );
            if( !(reply && reply.data()->parent ) ) break;

            parents.append( reply.data()->parent );
            if( reply.data()->parent == reply.data()->root ) break;
            else current = reply->parent;
        }

        if( !parents.empty() )
        {

            // store parent
            parent_ = parents.front();

            // store all parents in atom
            AtomData::Pointer atomData( new AtomData() );
            atomData->setName( "WINDOW_PARENTS" );
            atomData->setType( "WINDOW" );
            atomData->setFormat( 32 );
            atomData->setItemCount( parents.size() );
            atomData->setLength( parents.size() * sizeof( uint32_t ) );
            atomData->setData( &parents[0], parents.size() * sizeof( uint32_t ) );
            atomData->setValid( true );
            atoms_.insert( atomData->name(), atomData );
        }

    }

    // siblings
    if( !parents.empty() )
    {
        xcb_query_tree_cookie_t cookie = xcb_query_tree( connection, parents.front() );
        XcbUtil::ScopedPointer<xcb_query_tree_reply_t> reply(xcb_query_tree_reply( connection, cookie, 0x0 ) );
        if( reply )
        {
            xcb_window_t* siblings( xcb_query_tree_children( reply.data() ) );
            const int siblingCount( xcb_query_tree_children_length( reply.data() ) );

            // create atom
            if( siblingCount )
            {
                AtomData::Pointer atomData( new AtomData() );
                atomData->setName( "WINDOW_SIBLINGS" );
                atomData->setType( "WINDOW" );
                atomData->setFormat( 32 );
                atomData->setItemCount( siblingCount );
                atomData->setLength( siblingCount * sizeof( uint32_t ) );
                atomData->setData( siblings, siblingCount * sizeof( uint32_t ) );
                atomData->setValid( true );
                atoms_.insert( atomData->name(), atomData );
            }
        }

    }

    // children
    {
        xcb_query_tree_cookie_t cookie = xcb_query_tree( connection, window_ );
        XcbUtil::ScopedPointer<xcb_query_tree_reply_t> reply(xcb_query_tree_reply( connection, cookie, 0x0 ) );
        if( reply )
        {
            xcb_window_t* children( xcb_query_tree_children( reply.data() ) );
            const int childCount( xcb_query_tree_children_length( reply.data() ) );

            // create atom
            if( childCount )
            {
                AtomData::Pointer atomData( new AtomData() );
                atomData->setName( "WINDOW_CHILDREN" );
                atomData->setType( "WINDOW" );
                atomData->setFormat( 32 );
                atomData->setItemCount( childCount );
                atomData->setLength( childCount * sizeof( uint32_t ) );
                atomData->setData( children, childCount * sizeof( uint32_t ) );
                atomData->setValid( true );
                atoms_.insert( atomData->name(), atomData );
            }
        }

    }

    // parse all atoms
    for( AtomData::Hash::const_iterator iter = atoms_.constBegin(); iter != atoms_.constEnd(); ++iter )
    { iter.value()->parse( xcbProperties ); }

    // update validity
    valid_ = true;

    #endif

}

//_________________________________________________________
bool WindowProperties::isRootWindow( void ) const
{ return window_ == XcbUtil::get().appRootWindow(); }

//_____________________________________________________________________
QString WindowProperties::windowTitle( void ) const
{

    AtomData::Hash::const_iterator iter;
    if( (iter = atoms_.find( "_NET_WM_NAME" ) ) != atoms_.end() ) return iter.value()->stringValue();
    else if( (iter = atoms_.find( "WM_NAME" ) ) != atoms_.end() ) return iter.value()->stringValue();
    else return QString();

}

//_____________________________________________________________________
QString WindowProperties::windowClass( void ) const
{
    AtomData::Hash::const_iterator iter;
    if( ( iter = atoms_.find( "WM_CLASS" ) ) != atoms_.end() ) return iter.value()->stringValue();
    else return QString();

}

//_____________________________________________________________________
QString WindowProperties::command( void ) const
{

    AtomData::Hash::const_iterator iter;
    if( ( iter = atoms_.find( "WM_COMMAND" ) ) != atoms_.end() ) return iter.value()->stringValue();
    else return QString();

}

//_____________________________________________________________________
QString WindowProperties::processId( void ) const
{
    AtomData::Hash::const_iterator iter;
    if( (iter = atoms_.find( "_NET_WM_PID" ) ) != atoms_.end() ) return iter.value()->stringValue();
    else return QString();
}

//_____________________________________________________________________
QIcon WindowProperties::icon( void ) const
{
    QIcon out;
    AtomData::Hash::const_iterator iter( atoms_.find( "_NET_WM_ICON" ) );

    // check atom, validity and size
    if( iter == atoms_.end() || !iter.value()->isValid()|| iter.value()->format() != 32 ) return out;

    // pase data
    int offset( 0 );
    const uint32_t* data = reinterpret_cast<uint32_t*>( iter.value()->data() );
    while( true )
    {
        const int width( (int)data[offset] );
        const int height( (int)data[offset+1] );
        Debug::Throw() << "WindowProperties::icon: " << width << "x" << height << endl;
        const uint32_t* pixmapData = &data[offset+2];

        QImage image( width, height, QImage::Format_ARGB32);
        for( int i=0; i<image.byteCount()/4; ++i )
        { ((uint*)image.bits())[i] = pixmapData[i]; }

        out.addPixmap( QPixmap::fromImage( image ) );
        offset += 2+width * height;
        if( offset >= iter.value()->length()/iter.value()->format() ) break;
    }


    return out;
}
