// $Id: WindowPropertiesDisplay.cpp,v 1.9 2014/05/04 15:46:35 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "WindowPropertiesDisplay.h"
#include "WindowPropertiesDisplay.moc"

#include "AnimatedTreeView.h"
#include "GridLayout.h"
#include "GridLayoutItem.h"
#include "IconSize.h"
#include "IconSizeMenu.h"
#include "Singleton.h"
#include "XcbUtil.h"
#include "XmlOptions.h"

#include <QLayout>

//_________________________________________________________
WindowPropertiesDisplay::WindowPropertiesDisplay( QWidget* parent ):
    QWidget( parent ),
    Counter( "WindowPropertiesDisplay" )
{

    // horizontal layout
    QHBoxLayout* hLayout = new QHBoxLayout();
    hLayout->setSpacing(5);
    hLayout->setMargin(5);
    setLayout( hLayout );
    setSizePolicy( QSizePolicy::Minimum, QSizePolicy::Minimum );

    // icon label
    hLayout->addStretch( 1 );
    hLayout->addWidget( iconLabel_ = new QLabel( this ), 0 );

    // vLayout
    QVBoxLayout* vLayout = new QVBoxLayout();
    vLayout->setSpacing(5);
    vLayout->setMargin(0);
    hLayout->addLayout( vLayout, 10 );
    hLayout->addStretch( 1 );

    // grid layout
    GridLayout *gridLayout = new GridLayout();
    gridLayout->setSpacing(5);
    gridLayout->setMargin(0);
    gridLayout->setMaxCount(2);
    vLayout->addLayout( gridLayout, 0 );
    vLayout->addStretch( 1 );

    gridLayout->setColumnAlignment( 0, Qt::AlignVCenter|Qt::AlignRight );
    gridLayout->setColumnAlignment( 1, Qt::AlignVCenter|Qt::AlignLeft );

    ( nameItem_ = new GridLayoutItem( this, gridLayout, GridLayoutItem::All ) )->setKey( tr( "Window Title:" ) );
    ( classItem_ = new GridLayoutItem( this, gridLayout, GridLayoutItem::Elide|GridLayoutItem::Selectable ) )->setKey( tr( "Window Class:" ) );
    ( commandItem_ = new GridLayoutItem( this, gridLayout, GridLayoutItem::Elide|GridLayoutItem::Selectable ) )->setKey( tr( "Command:" ) );
    ( widItem_ = new GridLayoutItem( this, gridLayout, GridLayoutItem::Selectable ) )->setKey( tr( "Window Id:" ) );
    ( pidItem_ = new GridLayoutItem( this, gridLayout, GridLayoutItem::Selectable ) )->setKey( tr( "Process Id:" ) );

    separator_ = new QFrame( this );
    separator_->setFrameStyle( QFrame::HLine );
    gridLayout->addWidget( separator_, 5, 0, 1, 2 );

    ( sizeItem_ = new GridLayoutItem( this, gridLayout ) )->setKey( tr( "Window Size:" ) );
    ( positionItem_ = new GridLayoutItem( this, gridLayout ) )->setKey( tr( "Position:" ) );

    // icon size menu
    iconSizeMenu_ = new IconSizeMenu( this, true );
    connect( iconSizeMenu_, SIGNAL(iconSizeSelected(IconSize::Size)), SLOT(_updateIconSize(IconSize::Size)));

    // install menu for icon label context
    iconLabel_->setContextMenuPolicy( Qt::CustomContextMenu );
    connect( iconLabel_, SIGNAL(customContextMenuRequested(QPoint)), SLOT(_raiseIconSizeMenu(QPoint)) );

    // configuration update
    connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();

}

//_________________________________________________________
QString WindowPropertiesDisplay::windowClass( void ) const
{
    Debug::Throw( "WindowPropertiesDisplay::windowClass.\n" );
    return classItem_->text();
}

//_________________________________________________________
void WindowPropertiesDisplay::setWindowProperties( const WindowProperties& properties )
{

    // update icon
    icon_ = properties.icon();
    iconLabel_->setPixmap( icon_.isNull() ? QPixmap() : icon_.pixmap( IconSize( iconSize_ ) ) );

    // update items
    const WId windowId( properties.windowId() );
    if( windowId )
    {

        if( properties.isRootWindow() ) widItem_->setText( QString( "%1 (the root window)" ).arg( XcbUtil::get().windowIdString( windowId ) ) );
        else widItem_->setText( XcbUtil::get().windowIdString( windowId ) );

    } else widItem_->setText( QString() );

    nameItem_->setText( properties.windowTitle() );
    classItem_->setText( properties.windowClass() );
    commandItem_->setText( properties.command() );
    pidItem_->setText( properties.processId() );

    const QSize size( properties.size() );
    if( size.isValid() )
    {
        sizeItem_->setText( QString( "%1x%2" ).arg( size.width() ).arg( size.height() ) );

        const QPoint position( properties.position() );
        positionItem_->setText( QString( "(%1,%2)" ).arg( position.x() ).arg( position.y() ) );

    } else {

        sizeItem_->setText( QString() );
        positionItem_->setText( QString() );

    }

    _updateSeparatorVisibility();

}

//_________________________________________________________
void WindowPropertiesDisplay::_updateSeparatorVisibility( void )
{
    Debug::Throw() << "WindowPropertiesDisplay::_updateSeparatorVisibility" << endl;
    separator_->setVisible(
        ( nameItem_->isVisible() || classItem_->isVisible() || commandItem_->isVisible() || widItem_->isVisible() || pidItem_->isVisible() ) &&
        ( sizeItem_->isVisible() || positionItem_->isVisible() ) );
}

//_________________________________________________________
void WindowPropertiesDisplay::_raiseIconSizeMenu( QPoint position )
{
    Debug::Throw( "WindowPropertiesDisplay::_raiseIconSizeMenu.\n" );
    iconSizeMenu_->exec( iconLabel_->mapToGlobal( position ) );
}

//_________________________________________________________
void WindowPropertiesDisplay::_updateIconSize( IconSize::Size iconSize )
{
    Debug::Throw( "WindowPropertiesDisplay::_updateIconSize.\n" );

    // save value and store in options
    iconSize_ = iconSize;
    XmlOptions::get().set<int>( "WINDOW_ICON_SIZE", iconSize );

    // update label
    if( !icon_.isNull() )
    { iconLabel_->setPixmap( icon_.pixmap( IconSize( iconSize ) ) ); }

    return;

}

//_________________________________________________________
void WindowPropertiesDisplay::_updateConfiguration( void )
{
    Debug::Throw() << "WindowPropertiesDisplay::_updateConfiguration" << endl;
    iconSize_ = (IconSize::Size) XmlOptions::get().get<int>( "WINDOW_ICON_SIZE" );
    iconSizeMenu_->select( iconSize_ );
}
