// $Id: WindowSelectionFrame.cpp,v 1.23 2014/05/01 16:46:29 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "WindowSelectionFrame.h"
#include "WindowSelectionFrame.moc"

#include "LineEditor.h"
#include "IconNames.h"
#include "GridLayout.h"
#include "IconEngine.h"
#include "WarningDialog.h"
#include "Singleton.h"
#include "XmlOptions.h"
#include "XcbProperties.h"
#include "XcbUtil.h"

#include <QApplication>
#include <QLayout>
#include <QRegExpValidator>

//_________________________________________________________
WindowSelectionFrame::WindowSelectionFrame( QWidget* parent ):
    QWidget( parent ),
    Counter("WindowSelectionFrame" ),
    captureWidget_( 0 ),
    xcbProperties_( new XcbProperties( this ) ),
    mouseDown_( false )
{
    Debug::Throw( "WindowSelectionFrame::WindowSelectionFrame.\n" );

    QVBoxLayout* mainLayout = new QVBoxLayout();
    mainLayout->setMargin(5);
    mainLayout->setSpacing( 5 );
    setLayout( mainLayout );

    // select window
    QHBoxLayout* hLayout = new QHBoxLayout();
    hLayout->setSpacing( 5 );
    hLayout->setMargin( 0 );
    mainLayout->addLayout( hLayout );

    hLayout->addStretch(1);

    QLabel* label;
    hLayout->addWidget( label = new QLabel( tr( "Window Id:" ), this ) );
    hLayout->addWidget( editor_ = new LineEditor( this ) );
    editor_->setAlignment( Qt::AlignCenter );
    label->setBuddy( editor_ );
    editorLabel_ = label;
    editorLabel_->hide();
    editor_->setValidator( new QRegExpValidator( QRegExp( "((0x)((\\d)|[a-f]|[A-F])+)|(\\d+)" ), editor_ ) );
    editor_->hide();

    hLayout->addWidget( selectWindowButton_ = new QPushButton( this ), 0, Qt::AlignCenter );
    selectWindowButton_->setText( tr( "Select a New Window" ) );
    selectWindowButton_->setIcon( IconEngine::get( IconNames::Find ) );
    connect(selectWindowButton_, SIGNAL(clicked()), SLOT(selectWindow()));

    hLayout->addStretch(1);

    // separator
    QFrame* frame = new QFrame( this );
    frame->setFrameStyle( QFrame::HLine );
    mainLayout->addWidget( frame );

    // grid layout for options
    GridLayout* gridLayout( new GridLayout() );
    gridLayout->setSpacing( 5 );
    gridLayout->setMargin( 0 );
    gridLayout->setMaxCount( 2 );
    gridLayout->setColumnAlignment( 0, Qt::AlignVCenter|Qt::AlignRight );
    mainLayout->addLayout( gridLayout );

    // selection mode
    gridLayout->addWidget( label = new QLabel( tr( "Window selection mode:" ), this ) );
    gridLayout->addWidget( selectionModeComboBox_ = new QComboBox( this ) );
    selectionModeComboBox_->setEditable( false );
    selectionModeComboBox_->addItems( QStringList()
        << tr( "Window Under Cursor" )
        << tr( "Root Window" )
        << tr( "Window Id" ) );

    // selection delay
    gridLayout->addWidget( label = new QLabel( tr( "Window selection delay:" ), this ) );
    gridLayout->addWidget( delaySpinBox_ = new QSpinBox( this ) );
    label->setBuddy( delaySpinBox_ );
    delaySpinBox_->setSpecialValueText( tr( " No delay" ) );
    delaySpinBox_->setSuffix( tr( "s" ) );
    delaySpinBox_->setMinimum( 0 );
    delaySpinBox_->setMaximum( 99 );
    delayLabel_ = label;

    gridLayout->addWidget( label = new QLabel( tr( "Hide application when selecting a window:" ), this ) );
    gridLayout->addWidget( autoHideCheckBox_ = new QCheckBox( this ) );
    label->setBuddy( autoHideCheckBox_ );
    autoHideLabel_ = label;

    connect( editor_, SIGNAL(returnPressed()), SLOT(selectWindowFromEditor()) );
    connect( selectionModeComboBox_, SIGNAL(currentIndexChanged(int)), SLOT(_selectionModeChanged(int)) );

    // configuration update
    connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    connect( Singleton::get().application(), SIGNAL(saveConfiguration()), SLOT(_saveConfiguration()) );

    return;
}

//_________________________________________________________
bool WindowSelectionFrame::eventFilter( QObject* object, QEvent* event )
{

    // check object
    if( object != captureWidget_ ) return false;

    switch( event->type() )
    {
        case QEvent::MouseButtonPress:
        {
            QMouseEvent* mouseEvent( static_cast<QMouseEvent*>( event ) );
            if( mouseEvent->button() == Qt::LeftButton ) mouseDown_ = true;
            return true;
        }

        case QEvent::MouseMove:
        {
            if( mouseDown_ && _selectWindowFromPointer() && window_ )
            {
                editor_->setText( XcbUtil::get().windowIdString( window_ ) );
                reload();
            }
            return true;
        }

        case QEvent::MouseButtonRelease:
        {

            _clearCapture();

            if( autoHideCheckBox_->isChecked() )
            { window()->show(); }

            #if QT_VERSION >= 0x050000
            // need to explicitely release cursor for Qt5
            qApp->restoreOverrideCursor();
            #endif

            QMouseEvent* mouseEvent( static_cast<QMouseEvent*>( event ) );
            if( mouseEvent->button() == Qt::LeftButton )
            {
                _selectWindowFromPointer();
                editor_->setText( XcbUtil::get().windowIdString( window_ ) );
                reload();
            }

            return true;
        }

        default: return false;
    }
}



//_________________________________________________________
void WindowSelectionFrame::selectWindow(void)
{
    Debug::Throw( "WindowSelectionFrame::selectWindow" );
    switch( selectionModeComboBox_->currentIndex() )
    {
        case SelectFromPointer: return selectWindowFromPointer();
        case SelectRootWindow: return selectRootWindow();
        case SelectFromId: return selectWindowFromEditor();
        default: return;
    }

}

//_________________________________________________________
void WindowSelectionFrame::selectRootWindow(void)
{
    Debug::Throw( "WindowSelectionFrame::selectRootWindow" );
    #if HAVE_X11
    window_ = XcbUtil::get().appRootWindow();
    editor_->setText( XcbUtil::get().windowIdString( window_ ) );
    reload();
    #endif
}

//_________________________________________________________
void WindowSelectionFrame::selectWindowFromEditor( void )
{
    Debug::Throw( "WindowSelectionFrame::selectWindowFromEditor" );

    QString widString( editor_->text() );
    if( widString.startsWith( "0x" ) ) window_ = widString.mid( 2 ).toInt( 0, 16 );
    else window_ = widString.toInt();
    reload();

}

//_________________________________________________________
void WindowSelectionFrame::selectWindowFromTree( WId window )
{

    Debug::Throw( "WindowSelectionFrame::selectWindowFromTree.\n" );
    window_ = window;
    editor_->setText( XcbUtil::get().windowIdString( window_ ) );
    reload();

}

//_________________________________________________________
void WindowSelectionFrame::selectWindowFromPointer( bool forceImmediatly )
{
    Debug::Throw( "WindowSelectionFrame::selectWindowFromPointer.\n" );

    // auto hide
    if( autoHideCheckBox_->isChecked() ) window()->hide();

    const int delay( delaySpinBox_->value() );
    if( delay > 0 && !forceImmediatly )
    {

        // hide main application and start timer
        timer_.start( delay*1000, this );

    } else {

        _clearCapture();
        captureWidget_ = new QDialog( this, Qt::X11BypassWindowManagerHint );
        captureWidget_->installEventFilter( this );
        captureWidget_->move( -1000, -1000 );
        captureWidget_->setModal( true );
        captureWidget_->show();
        captureWidget_->grabMouse( Qt::CrossCursor );

        #if QT_VERSION >= 0x050000
        // need to explicitely override cursor for Qt5
        qApp->setOverrideCursor( Qt::CrossCursor );
        #endif

    }

    return;
}

//_________________________________________________________
void WindowSelectionFrame::reload( void )
{

    Debug::Throw( "WindowSelectionFrame::reload.\n" );

    // check window
    if( !window_ )
    {
        WarningDialog( this, tr( "Selected window is invalid." ) ).exec();
        return;
    }

    // get properties and check
    windowProperties_ = WindowProperties( window_, xcbProperties_ );
    if( !windowProperties_.isValid() )
    {
        WarningDialog( this, QString( tr( "Window %1 is invalid." ) ).arg( XcbUtil::get().windowIdString( window_ ) ) ).exec();
        return;
    }

    // emit signal changed
    emit windowSelectionChanged();
    return;
}

//_____________________________________________________________
void WindowSelectionFrame::timerEvent( QTimerEvent *event )
{

    Debug::Throw( "WindowSelectionFrame::timerEvent.\n" );

    if( event->timerId() == timer_.timerId() )
    {
        timer_.stop();

        // find window
        _selectWindowFromPointer();

        // show this application
        if( autoHideCheckBox_->isChecked() )
        { window()->show(); }

        // update window properties
        reload();

        return;
    } else return QWidget::timerEvent( event );

}

//_________________________________________________________
bool WindowSelectionFrame::_selectWindowFromPointer( void )
{

    Debug::Throw( "WindowSelectionFrame::_selectWindowFromPointer.\n" );

    // reset window
    window_ = 0L;

    #if HAVE_X11
    xcb_connection_t* connection( XcbUtil::get().connection<xcb_connection_t>() );
    xcb_query_pointer_cookie_t cookie( xcb_query_pointer( connection, XcbUtil::get().appRootWindow() ) );
    XcbUtil::ScopedPointer<xcb_query_pointer_reply_t> reply( xcb_query_pointer_reply( connection, cookie, 0x0 ) );
    if( !reply ) return false;

    WId window( reply->child ? reply->child:reply->root );
    WId realWindow( _findRealWindow( window ) );
    if( realWindow ) window = realWindow;
    if( window_ != window )
    {
        window_ = window;
        return true;
    }
    #endif

    return false;
}

//_________________________________________________________
WId WindowSelectionFrame::_findRealWindow( WId window, int depth )
{

    #if HAVE_X11

    // check recursion level
    if( depth > 5 ) return 0L;
    else if( XcbUtil::get().isRealWindow( window ) ) return window;
    else {

        // parse children
        xcb_connection_t* connection( XcbUtil::get().connection<xcb_connection_t>() );
        xcb_query_tree_cookie_t cookie = xcb_query_tree( connection, window );
        XcbUtil::ScopedPointer<xcb_query_tree_reply_t> reply( xcb_query_tree_reply( connection, cookie, 0x0 ) );

        WId out = 0L;
        if( reply )
        {

            const int childCount( xcb_query_tree_children_length( reply.data() ) );
            xcb_window_t* children( xcb_query_tree_children( reply.data() ) );
            for( int i=0; i < childCount && !out; ++i )
            { out = _findRealWindow( children[i], depth+1 ); }

        }

        return out;

    }

    #else

    return 0L;

    #endif

}

//_________________________________________________________
void WindowSelectionFrame::_clearCapture( void )
{
    if( captureWidget_ )
    {

        Debug::Throw( "WindowSelectionFrame::_clearCapture.\n" );

        // release mouse and delete widget
        captureWidget_->releaseMouse();
        captureWidget_->deleteLater();
        captureWidget_ = 0;
    }

    mouseDown_ = false;
}

//_________________________________________________________
void WindowSelectionFrame::_selectionModeChanged( int value )
{

    Debug::Throw() << "WindowSelectionFrame::_selectionModeChanged - value: " << value << endl;

    switch( value )
    {
        default:
        case SelectFromPointer:
        selectWindowButton_->setText( tr( "Select a New Window" ) );
        selectWindowButton_->setIcon( IconEngine::get( IconNames::Find ) );
        delaySpinBox_->setEnabled( true );
        delayLabel_->setEnabled( true );
        autoHideCheckBox_->setEnabled( true );
        autoHideLabel_->setEnabled( true );

        editorLabel_->hide();
        editor_->hide();
        break;

        case SelectRootWindow:
        selectRootWindow();
        selectWindowButton_->setText( tr( "Update" ) );
        selectWindowButton_->setIcon( IconEngine::get( IconNames::Reload ) );

        delaySpinBox_->setEnabled( false );
        delayLabel_->setEnabled( false );
        autoHideCheckBox_->setEnabled( false );
        autoHideLabel_->setEnabled( false );

        editorLabel_->hide();
        editor_->hide();
        break;

        case SelectFromId:
        selectWindowButton_->setText( tr( "Update" ) );
        selectWindowButton_->setIcon( IconEngine::get( IconNames::Reload ) );
        delaySpinBox_->setEnabled( false );
        delayLabel_->setEnabled( false );
        autoHideCheckBox_->setEnabled( false );
        autoHideLabel_->setEnabled( false );

        editorLabel_->show();
        editor_->show();
        break;
    }

}

//_________________________________________________________
void WindowSelectionFrame::_updateConfiguration( void )
{
    Debug::Throw( "WindowSelectionFrame::_updateConfiguration.\n" );
    delaySpinBox_->setValue( XmlOptions::get().get<int>( "GRAB_DELAY" ) );
    autoHideCheckBox_->setChecked( XmlOptions::get().get<bool>( "AUTO_HIDE" ) );
    selectionModeComboBox_->setCurrentIndex( XmlOptions::get().get<int>( "SELECTION_MODE" ) );
}

//_________________________________________________________
void WindowSelectionFrame::_saveConfiguration( void )
{
    Debug::Throw( "WindowSelectionFrame::_saveConfiguration.\n" );
    XmlOptions::get().set<int>( "GRAB_DELAY", delaySpinBox_->value() );
    XmlOptions::get().set<bool>( "AUTO_HIDE", autoHideCheckBox_->isChecked() );
}
