
// $Id: WindowTreeWidget.cpp,v 1.12 2014/05/09 07:21:28 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "WindowTreeWidget.h"
#include "WindowTreeWidget.moc"

#include "ContextMenu.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "Singleton.h"
#include "TreeView.h"
#include "XcbUtil.h"
#include "XcbWindowTreeParser.h"
#include "XmlOptions.h"

#include <QHeaderView>
#include <QLayout>

//__________________________________________________
WindowTreeWidget::WindowTreeWidget( QWidget* parent ):
    QWidget( parent ),
    Counter( "WindowTreeWidget" ),
    windowTreeParser_( new XcbWindowTreeParser( this ) ),
    defaultWidth_( -1 )
{

    Debug::Throw( "WindowTreeWidget::WindowTreeWidget.\n" );

    // actions
    _installActions();

    QVBoxLayout* layout = new QVBoxLayout();
    layout->setMargin(0);
    layout->setSpacing(5);
    setLayout( layout );

    layout->addWidget( list_ = new TreeView( this ) );
    list_->setRootIsDecorated( true );
    list_->setSortingEnabled( true );
    list_->setModel( &model_ );
    list_->setOptionName( "WINDOW_TREE_LIST" );

    connect( list_, SIGNAL(activated(QModelIndex)), SLOT(_indexActivated(QModelIndex)) );
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateActions()) );

    // menu
    QMenu* menu = new ContextMenu( list_ );
    menu->addAction( showAllWindowsAction_ );
    menu->addAction( reloadAction_ );
    menu->addAction( selectWindowAction_ );

    // configuration update
    connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    _updateConfiguration();

    // root
    rootWindowId_ = WindowTreeItem( XcbUtil::get().appRootWindow(), 0x0, WindowTreeItem::Root );
    model_.add( rootWindowId_ );
    list_->setExpanded( model_.index( rootWindowId_ ), true );

}

//______________________________________________________________________
void WindowTreeWidget::setDefaultWidth( int value )
{ defaultWidth_ = value; }

//____________________________________________
QSize WindowTreeWidget::sizeHint( void ) const
{ return (defaultWidth_ ) >= 0 ? QSize( defaultWidth_, 0 ):QWidget::sizeHint(); }

//____________________________________________
void WindowTreeWidget::setVisible( bool value )
{
    XmlOptions::get().set<bool>( "SHOW_WINDOW_TREE", value );
    QWidget::setVisible( value );
}

//__________________________________________________
void WindowTreeWidget::selectWindow( WindowTreeItem item )
{
    Debug::Throw( "WindowTreeWidget::selectWindow.\n" );
    QModelIndex index( model_.index( item ) );
    if( !index.isValid() ) return;

    list_->selectionModel()->select( index, QItemSelectionModel::Clear|QItemSelectionModel::Select|QItemSelectionModel::Rows );
    list_->selectionModel()->setCurrentIndex( index, QItemSelectionModel::Current );
    list_->scrollTo( index );

}

//__________________________________________________
void WindowTreeWidget::reload( void )
{

    Debug::Throw( "WindowTreeWidget::reload.\n" );

    // store selected jobs in model
    list_->setUpdatesEnabled( false );
    list_->storeScrollBarPosition();
    list_->storeSelectedIndexes();
    list_->storeExpandedIndexes();

    model_.set( windowTreeParser_->parse( rootWindowId_ ) );

    // restore selected jobs
    list_->resizeColumns();
    list_->restoreExpandedIndexes();
    list_->restoreSelectedIndexes();
    list_->restoreScrollBarPosition();
    list_->setUpdatesEnabled( true );


}

//__________________________________________________
void WindowTreeWidget::reloadOnWindowChanged( void )
{
    if( !timer_.isActive() )
    {
        Debug::Throw() << "WindowTreeWidget::reloadOnWindowChanged." << endl;
        timer_.start( 200, this );
    }
}

//__________________________________________________
void WindowTreeWidget::timerEvent( QTimerEvent* event )
{
    if( event->timerId() == timer_.timerId() )
    {
        timer_.stop();
        reload();
    } else return QWidget::timerEvent( event );
}

//__________________________________________________
void WindowTreeWidget::_toggleShowAllWindows( bool value )
{
    Debug::Throw( "WindowTreeWidget::_toggleShowAllWindows.\n" );
    XmlOptions::get().set<bool>( "SHOW_ALL_WINDOWS", value );

    model_.setShowAllWindows( value );
    windowTreeParser_->setShowAllWindows( value );
    reload();

}

//__________________________________________________
void WindowTreeWidget::_indexActivated( QModelIndex index )
{
    Debug::Throw( "WindowTreeWidget::_indexActivated.\n" );
    WindowTreeItem item( model_.get( index ) );
    emit windowActivated( item.id() );
}

//__________________________________________________
void WindowTreeWidget::_updateActions( void )
{
    Debug::Throw( "WindowTreeWidget::_showWindowProperties.\n" );
    selectWindowAction_->setEnabled( list_->selectionModel()->hasSelection() );
}

//__________________________________________________
void WindowTreeWidget::_showWindowProperties( void )
{
    Debug::Throw( "WindowTreeWidget::_showWindowProperties.\n" );
    if( !list_->selectionModel()->hasSelection() ) return;
    const QModelIndex index( list_->selectionModel()->selectedIndexes().front() );
    if( index.isValid() )
    { emit windowActivated( model_.get( index ).id() ); }
}

//_________________________________________________________
void WindowTreeWidget::_updateConfiguration( void )
{
    Debug::Throw() << "WindowTreeWidget::_updateConfiguration" << endl;
    showAllWindowsAction_->setChecked( XmlOptions::get().get<bool>( "SHOW_ALL_WINDOWS" ) );
    visibilityAction_->setChecked( XmlOptions::get().get<bool>( "SHOW_WINDOW_TREE" ) );
}

//_________________________________________________________
void WindowTreeWidget::_installActions( void )
{
    Debug::Throw() << "WindowTreeWidget::_installActions." << endl;

    addAction( visibilityAction_ = new QAction( tr( "Show Window Tree" ), this ) );
    visibilityAction_->setCheckable( true );
    visibilityAction_->setChecked( true );
    connect( visibilityAction_, SIGNAL(toggled(bool)), SLOT(setVisible(bool)) );

    addAction( showAllWindowsAction_ = new QAction( tr( "Show All Windows" ), this ) );
    showAllWindowsAction_->setCheckable( true );
    showAllWindowsAction_->setChecked( false );
    connect( showAllWindowsAction_, SIGNAL(toggled(bool)), SLOT( _toggleShowAllWindows(bool)) );

    addAction( reloadAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Reload Window Tree" ), this ) );
    connect( reloadAction_, SIGNAL(triggered()), SLOT(reload()) );

    addAction( selectWindowAction_ = new QAction( IconEngine::get( IconNames::Find ), tr( "View Window Properties" ), this ) );
    connect( selectWindowAction_, SIGNAL(triggered()), SLOT(_showWindowProperties()) );
    selectWindowAction_->setEnabled( false );

    connect( visibilityAction_, SIGNAL(toggled(bool)), showAllWindowsAction_, SLOT(setEnabled(bool)) );
    connect( visibilityAction_, SIGNAL(toggled(bool)), reloadAction_, SLOT(setEnabled(bool)) );

}
