// $Id: XcbWindowTreeParser.cpp,v 1.8 2014/05/08 10:15:38 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "XcbWindowTreeParser.h"
#include "XcbWindowTreeParser.moc"

#include "Debug.h"
#include "XcbProperties.h"
#include "XcbUtil.h"

#if HAVE_X11
#include <xcb/xcb.h>
#endif

//____________________________________________________________________
XcbWindowTreeParser::XcbWindowTreeParser( QObject* parent ):
    QObject( parent ),
    Counter( "XcbWindowTreeParser" ),
    xcbProperties_( new XcbProperties( this ) ),
    foundRoot_( false ),
    showAllWindows_( false )
{}

//____________________________________________________________________
WindowTreeItem::List XcbWindowTreeParser::parse( WindowTreeItem window )
{
    foundRoot_ = false;
    return _parse( window );
}

//____________________________________________________________________
WindowTreeItem::List XcbWindowTreeParser::_parse( WindowTreeItem window )
{

    WindowTreeItem::List out;

    #if HAVE_X11
    // adjust item flags
    if( !foundRoot_ )
    {
        if( window.id() == XcbUtil::get().appRootWindow() )
        { window.setFlag( WindowTreeItem::Root, true ); }
        foundRoot_ = true;
    }

    if( XcbUtil::get().isRealWindow( window.id() ) )
    {
        window.setFlag( WindowTreeItem::Real, true );
    }

    window.setWindowName( _windowName( window.id() ) );

    // append to list
    if( showAllWindows_ || (window.flags() & (WindowTreeItem::Real|WindowTreeItem::Root) ) )
    { out.append( window ); }

    // parse children
    xcb_connection_t* connection( XcbUtil::get().connection<xcb_connection_t>() );
    xcb_query_tree_cookie_t cookie = xcb_query_tree( connection, window.id() );
    XcbUtil::ScopedPointer<xcb_query_tree_reply_t> reply(xcb_query_tree_reply( connection, cookie, 0x0 ) );
    if( !reply ) return out;

    xcb_window_t* children( xcb_query_tree_children( reply.data() ) );
    const int childCount( xcb_query_tree_children_length( reply.data() ) );
    for( int iChild = 0; iChild < childCount; ++iChild )
    { out.append( _parse( WindowTreeItem( children[iChild], window.id() ) ) ); }

    #endif

    return out;

}

//____________________________________________________________________
QString XcbWindowTreeParser::_windowName( WId window ) const
{
    Debug::Throw() << "XcbWindowTreeParser::_windowName - " << XcbUtil::get().windowIdString( window ) << endl;
    #if HAVE_X11

    // try using _NET_WM_NAME
    if( XcbUtil::get().isSupported( X11Defines::_NET_WM_NAME ) )
    {

        // get WM_NAME atom
        xcb_atom_t& atom( *XcbUtil::get().atom<xcb_atom_t>( X11Defines::_NET_WM_NAME ) );

        AtomData::Pointer atomData( xcbProperties_->atomData( window, atom ) );
        if( atomData->isValid() ) 
        {
            atomData->parse( xcbProperties_ ); 
            return atomData->stringValue();
        }
        
    }
    
    // try using WM_NAME
    xcb_atom_t& atom( *XcbUtil::get().atom<xcb_atom_t>( "WM_NAME" ) );
    if( atom )
    {
        AtomData::Pointer atomData( xcbProperties_->atomData( window, atom ) );
        if( atomData->isValid() && atomData->type() == "STRING" ) 
        {
            atomData->parse( xcbProperties_ ); 
            return atomData->stringValue();
        }
    }
    
    #endif
    
    return QString();

}
